

Make

Author: Stuart Feldman

Make originated with a visit from Steve Johnson (author of yacc, etc.), 
storming into my office, cursing the Fates that had caused him to waste 
a morning debugging a correct program (bug had been fixed, file had not
been compiled, cc *.o was therefore unaffected). As I had spent a part
of the previous evening coping with the same disaster on a project I was
working on, the idea of a tool to solve it came up. It began with an
elaborate idea of a dependency analyzer, boiled down to something much
simpler, and turned into Make that weekend. Use of tools that were still
wet was part of the culture. Makefiles were text files, not magically
encoded binaries, because that was the Unix ethos: printable, debuggable,
understandable stuff.
 	
-- Stuart Feldman 	 



.h --->
        .s ----> .o ----> a.out
.c --->       


make components
 - explicit rules
 - implicit rules
 - variable definitions
 - dependencies 
 - productions
 - comments

explicit rule 
    how to remake the target

implicit rule
    how to remake a class of files based on their names

variable definitions
    src = a.cpp b.cpp \
          c.cpp d.cpp

directives
    reading other makefiles
        include otherMakefile1 otherMakefile2 ...

    (Trick: generate depenences into otherMakefile, than execute it)    

    decide whether to ignore some part of a makefile
    defining variable from a verbatim string

comments
    #



Wildcards - like in shell

    '*'  '?'  '[...]' 


Automatic variables

    $@  the file name of the target of the rule

    $%  the tarhet member name, when the target i san archive member 

    $<  the name of the first prerequisite

    $?  the name of all prerequisits that are newer than the target
    
    $^  the names of all prerequisits (separated by spaces)
    
    $+  similar than the previous

    $*  

    $(@D)   the directory part of the target

    $(@F)   the file part of the target
    

Conditional parts

    ifeq($(CC),gcc)
        $(CC) -o foo $(objects) $(libs_for_gcc)
    else
        $(CC) -o foo $(objects) $(libs_for_other)
    endif


Implicit rules

    foo:  foo.o bar.o
        cc -o foo foo.o bar.o $(CFLAGS) $(LDFLAGS)

    No rule for foo.o, but make automatically looks for it.       


    x.o : x.c
        $(CC) -c $(CPPFLAGS) $(CFLAGS) x.c
        
    x.o : x.cc | x.C 
        $(CXX) -c $(CPPFLAGS) $(CXXFLAGS) x.cc x.C

    Also there are rules for pascal, fortran, modula2, assembly

    Linking a single object:

    x: x.o
        $(CC) $(LDFLAGS) x.o $(LOADLIBES) $(LDLIBS)


We can define custom implicite rules by writing a pattern rule

    Looks like an ordinary rule, but the target includes exactly 
    one '%' character.

    %.c  as a pattern matches any file name that ends .c
    s.%.c matches any at least 5 char strings starting 's' ending 'c'

    The actual matching is called the "stem".

    The '%' in the prerequisite of a pattern rule stands for the 
    same stem that was matched by the '%' in the target.

    %.o : %.c ; command ...

    Sometimes the prerequisites do not use '%', matching all files. 

    Examples:

    %.o : %.c
        $(CC) -c $(CFLAGS) $(CPPFLAGS) $< -o $@

        $@ is the target and $< is the source 
        for each case the rule applies.        

    % :: RCS/%,v
        $(CO) $(COFLAGS) $<
        
        for all files in RCS subdirectory     
        :: is a terminal rule.

Suffix rules (old style)

    .o : .c     is equivalent with      %.o : %.c

    .c.o : 
        $(CC) -c $(CFLAGS) $(CPPFLAGS) -o $@ $<
                                                    is today:
    %.c : %.o
        $(CC) -c $(CFLAGS) $(CPPFLAGS) -o $@ $<
    

Functions

    $(function arguments...)
    ${function arguments...}
    


    $(subst from,to,text)
    $(strip string)
    $(findstring what,where) 
    $(sort list)
    $(word n text)
    $(dir names...)
    $(suffix names...)
    $(wildcard pattern...)
    

    Wildcard expansion happens automatically in rules. But no otherwise.    

    Examples:

        $(wildcard *.c)     lists all the .c files
        $(patsubst %.c,%.o,$(wildcard *.c))  changing x.c to x.o 

Very complex makefiles, especially when they call subsidiary makefiles,
can become a source of complications rather than simplifying the build
process. A now-classic warning is issued in Recursive Make Considered
Harmful.
http://www.tip.net.au/~millerp/rmch/recu-make-cons-harm.html

Lines starting vith TABs - command lines - are presented to the shell
as they are - including the TABs.

Why the tab in column 1? Yacc was new, Lex was brand new. I had not tried 
either, so I figured this would be a good excuse to learn. After getting
myself snarled up with my first stab at Lex, I just did something simple 
with the pattern newline-tab. It worked, it stayed. And then a few weeks
later I had a user population of about a dozen, most of them friends, and
I did not want to screw up my embedded base. The rest, sadly, is history.
 	
-- Stuart Feldman 	 


    $ make
        looks for filenames, like GNUmakefile makefile and Makefile

    $ make -f my_make_file.mk                  
vagy
    $ make --file 

If a MAKEFILES environment variable is exists, then make read
makefiles enlisted in MAKEFILES.

MAKEFILE_LIST is automatically appended by make with the names
of the executed makefiles.

.VARIABLES
    the names defined by make

    $ make -n       # only prints the commands not executes.


Typical non-file (.PHONY) productions:

all

- this should be the first production in your makefile. 
- executed when the developer types make with no argument.

test

- run the automated test suit or check installation.

clean

- remove all files created by make all.

dist

- make a source archive can be shipped as a unit 
- equivalent of depending on all

install
uninstall


######################## sample ######################

.PHONY: doc clean show graph 

SRC	= $(wildcard src/*)
TEXSRC	= $(wildcard *.tex)
DIASRC  = $(wildcard pic/*.dia)

article.dvi: $(TEXSRC) $(DIASRC)

%.dvi: %.tex
	latex $<
	latex $<

%.eps: pic/%.dia
	dia --nosplash --export=$@ $<


default: doc

show: doc
	xdvi.bin article.dvi &

doc: pics $(SRC) article.dvi

pics: $(DIASRC) 

purge: clean
	rm -f *.eps
        
clean:
	rm -f *.aux
	rm -f *.lo*
	rm -f *.toc
	rm -f *.dvi
	rm -f *.bbl *.blg
	rm -f *~ pic/*~ src/*~

help:
	@echo "Available targets:"
	@echo "  show	   - compile and show document"
	@echo "  doc	   - compile documents (default)"
	@echo "  pics      - compile pictures to document format"
	@echo "  clean	   - remove generated files"
#	@echo "  retab     - retabulates the sources"
#	@echo "  longlines - show too long lines in the sources"
	@echo ""


# --- Still unused
%.eps: %.plot
	echo 'set terminal postscript eps; set output "$@"' | \
		gnuplot - $<

retab:
	for i in $(SRC); do vim -s retab.vim -e $$i; done

longlines:
	egrep '^.{76,}' $(SRC)


#source.tex: $(SRC)
#graph: benchmark.eps

######################### end sample #########################


******************************************************************


Makefile generators:


makedepend
==========

- reads sources, parses precompiler directives

SRCS = file1.c file2.c ...
CFLAGS = -O -DHACK -I../foobar -xyz
depend:
    makedepend -- $(CFLAGS) -- $(SRCS)
                                                                                

// a.h:
#include "b.h"
#include "b.c"

// b.h:
/* empty */

// a.c:
#include <list>
#include "a.h"

int main()
{
    return 0;
}

// b.c:
#include "b.h"


$ makedepend -fMakefile a.c
$ cat Makefile

# DO NOT DELETE
 
a.o: a.h b.h
b.o: b.h



*****************************************************************




Imake
=====

Imake is a Makefile generator to separate machine-dependent issues
from other make instructions.

Imake is based on CPP (C precompiler).



Imake was written in an attempt to mechanize makefile generation
for the X window system. It builds on makedepend to tackle both 
the dependency-derivation and portability problems.

hard to learn, hard to use  :(

Archive for "Software Portability with imake" (2nd edition) (O Reilly)
http://www.kitebird.com/imake-book/
FAQ:
http://www.snake.net/software/imake-stuff/imake-faq.html

$ xmkmf
mv -f Makefile Makefile.bak
imake -DUseInstalled -I/usr/X11R6/lib/X11/config
imake: No such file or directory
imake: No description file.
  Stop.


You never run imake directly. You should use a bootstrapper, to tell
imake 


********************************************************************



autoconf
========

- produces per-project configure shell scripts
- shell scripts generate makefiles

Manual:
http://www.gnu.org/software/autoconf/manual/index.html


your source files --> [autoscan*] --> [configure.scan] --> configure.ac
configure.ac --.
               |   .------> autoconf* -----> configure
[aclocal.m4] --+---+
               |   .-----> [autoheader*] --> [config.h.in]
[acsite.m4] ---.

Makefile.in -------------------------------> Makefile.in


Files used in configuring a software package:

                         .-------------> [config.cache]
configure* --------------+-------------> config.log
                         |
[config.h.in] ----.      v             .-> [config.h] -.
                  +--> config.status* -+               +--> make*
Makefile.in ------.                    .-> Makefile ---.  


automake
========

- put imake-like dependency derivation as a layer on 
  top of autoconf. 
- relatively new

              automake
Makefile.am  ---------> Makefile.in 



$ make love
  "Don't know how to make love"                     (original)
  make: *** No rule to make target "love".  Stop.   (current gnu make)

