#ifndef PQUEUE_H
#define PQUEUE_H

#include <utility>
#include <list>
#include <vector>
#include <algorithm>

template <typename T, typename P>
class pqueue
{
public:
    pqueue& push(const T& t, const P& p); 
    T       pop(); 

    int  size()  const { return lst.size(); }
    bool empty() const { return lst.empty(); }

    bool has( const T& t) const;    
    int  count(const P& p) const;    
    P    max() const;
    void getprior(const P& p, std::vector<T>& v) const;
private:
    std::list< std::pair<T,P> > lst;
};
template <typename T, typename P>
void pqueue<T,P>::getprior(const P& p, std::vector<T>& v) const
{
    typename std::list< std::pair<T,P> >::const_iterator ci = lst.begin();

    while ( ci != lst.end() )
    {   
        if ( ci->second == p )
            v.push_back(ci->first);
        ++ci;
    }
}
template <typename T, typename P>
pqueue<T,P>& pqueue<T,P>::push(const T& t, const P& p)
{
    std::pair<T,P> pr(t,p);
    lst.push_back( pr );
    return *this;
}
template <typename T, typename P>
bool pqueue<T,P>::has(const T& t) const
{
    typename std::list< std::pair<T,P> >::const_iterator ci = lst.begin();

    while ( ci != lst.end() )
    {   
        if ( ci->first == t )
            return true;
        ++ci;
    }
    return false;
}
template <typename T, typename P>
int pqueue<T,P>::count(const P& p) const
{
    int cnt = 0;
    typename std::list< std::pair<T,P> >::const_iterator ci = lst.begin();
    while ( ci != lst.end() )
    {   
        if ( ci->second == p )
            ++cnt;
        ++ci;
    }
    return cnt;
}
template <typename T, typename P>
P pqueue<T,P>::max() const
{
    typename std::list< std::pair<T,P> >::const_iterator ci = lst.begin();
    typename std::list< std::pair<T,P> >::const_iterator cj = ci;

    while ( ci != lst.end() )
    {   
        if ( ci->second > cj->second )
            cj = ci;
        ++ci;
    }
    return cj->second;
}
template <typename T, typename P>
T pqueue<T,P>::pop()
{
     
    typename std::list< std::pair<T,P> >::iterator i = lst.begin();
    typename std::list< std::pair<T,P> >::iterator j = i;

    while ( i != lst.end() )
    {   
        if ( i->second > j->second )
            j = i;
        ++i;
    }
    T t = j->first;
    lst.erase(j);
    return t;
}


#endif /* PQUEUE_H */

