#ifndef ESET_H
#define ESET_H


#include <vector>
#include <algorithm>

template <typename T>
class eset
{
public:
    eset& put( const T &t);
    eset& del( const T &t);
    bool  has( const T &t) const;

    bool  is_empty() const;
    int   size() const;
    size_t  max_size() const { return vec.max_size(); }
    void  print( std::ostream &os) const;

    eset& operator+=( const T &t);
    eset& operator+=( const eset<T> &t);
private:
    std::vector<T> vec;
}; 

template <typename T>
void eset<T>::print( std::ostream &os) const
{
    for( vector<int>::const_iterator ci = vec.begin(); ci != vec.end(); ++ci)
        os << *ci << ", ";
}

template <typename T>
eset<T>& eset<T>::operator+=( const T &t)
{
    put(t);
    return *this;
}

template <typename T>
eset<T>& eset<T>::operator+=( const eset<T> &t)
{
    vec.insert(vec.begin(), t.vec.begin(), t.vec.end());
    sort(vec.begin(), vec.end());
    typename vector<T>::iterator e = unique(vec.begin(), vec.end());
    vec.erase( e, vec.end());
    return *this;
}

template <typename T>
eset<T> operator+( const eset<T> &t1,  const eset<T> &t2)
{
    eset<T> result = t1;
    result += t2;
    return result;
}

template <typename T>
eset<T>& eset<T>::put( const T &t)
{
    typename vector<T>::iterator i = find( vec.begin(), vec.end(), t);
    
    if ( vec.end() == i )   vec.push_back(t); 
    return *this;
}

template <typename T>
eset<T>& eset<T>::del( const T &t)
{
    typename vector<T>::iterator i = find( vec.begin(), vec.end(), t);
    
    if ( vec.end() != i )   vec.erase(i); 
    return *this;
}

template <typename T>
bool eset<T>::has( const T &t) const
{
    return vec.end() != find( vec.begin(), vec.end(), t);
}

template <typename T>
bool eset<T>::is_empty() const
{
    return 0 == size();
}

template <typename T>
int eset<T>::size() const
{
    return vec.size();
}
    

#endif /* ESET_H */
