#ifndef SET_H
#define SET_H


#include "eset.h"


template <typename T>
class set
{
public:
    set() : c(false) { }

    set& put( const T &t);
    set& del( const T &t);
    bool has( const T &t) const;

    bool is_empty() const;
    size_t size() const;
    void print( std::ostream &os) const;

    set& operator+=( const T &t);
    set& operator-=( const T &t);
    set  operator~() const;
private:
    bool    c;
    eset<T> e;
}; 

template <typename T>
set<T>& set<T>::put( const T &t)
{
    if ( !c )
        e.put(t);
    else
        e.del(t);
    return *this;
}

template <typename T>
set<T>& set<T>::del( const T &t)
{
    if ( !c )
        e.del(t);
    else
        e.put(t);
    return *this;
}

template <typename T>
bool set<T>::has( const T &t) const
{
    if ( !c )
        return e.has(t);
    else
        return !e.has(t);
}

template <typename T>
bool set<T>::is_empty() const
{
    if ( !c )
        return e.is_empty();
    else
        return !e.is_empty();
}

template <typename T>
size_t set<T>::size() const
{
    if ( !c )
        return e.size();
    else
        return e.max_size() - e.size();
}

template <typename T>
set<T> set<T>::operator~() const
{
    set<T> result(*this);
    result.c = !result.c;
    return result;
}

template <typename T>
set<T>& set<T>::operator+=( const T &t)
{
    put(t);
    return *this;
}

template <typename T>
set<T>& set<T>::operator-=( const T &t)
{
    del(t);
    return *this;
}

template <typename T>
void set<T>::print(std::ostream &os) const
{
    os << " { ";
    if ( c )
        os << "not in: ";
    e.print(os);
    os << "} ";
}


#endif /* SET_H */

